"""Module for the Simplex Domain."""

from copy import deepcopy
import torch
from .base_domain import BaseDomain
from ..label_tensor import LabelTensor
from ..utils import check_consistency


class SimplexDomain(BaseDomain):
    """
    Implementation of the simplex domain.

    :Example:

        >>> simplex_domain = SimplexDomain(
                [
                    LabelTensor(torch.tensor([[0, 0]]), labels=["x", "y"]),
                    LabelTensor(torch.tensor([[1, 1]]), labels=["x", "y"]),
                    LabelTensor(torch.tensor([[0, 1]]), labels=["x", "y"]),
                ]
            )
    """

    def __init__(self, simplex_matrix, sample_surface=False):
        """
        Initialization of the :class:`SimplexDomain` class.

        :param simplex_matrix: The matrix of the simplex vertices.
        :type simplex_matrix: list[LabelTensor] | tuple[LabelTensor]
        :param bool sample_surface: If ``True``, only the surface of the simplex
            is considered part of the domain. Default is ``False``.
        :raises ValueError: If any element of ``simplex_matrix`` is not a
            :class:`LabelTensor`.
        :raises TypeError: If ``simplex_matrix`` is not a list or tuple.
        :raises ValueError: If ``sample_surface`` is not a boolean.
        :raises ValueError: If the labels of the vertices do not match.
        :raises ValueError: If the number of vertices is not equal to the
            dimension of the simplex plus one.
        """
        super().__init__()

        # Initialization
        self._sample_modes = ("random",)
        self.sample_surface = sample_surface
        self.vert_matrix = simplex_matrix

    def is_inside(self, point, check_border=False):
        """
        Check if a point is inside the simplex.

        :param LabelTensor point: The point to check.
        :param bool check_border: If ``True``, the boundary is considered inside
            the domain. Default is ``False``.
        :raises ValueError: If ``point`` is not a :class:`LabelTensor`.
        :raises ValueError: If the labels of ``point`` differ from the variables
            of the domain.
        :return: Whether the point is inside the domain or not.
        :rtype: bool
        """
        # Checks on point
        check_consistency(point, LabelTensor)
        if set(self.variables) != set(point.labels):
            raise ValueError(
                "Point labels differ from constructor vertices labels. "
                f"Got {sorted(point.labels)}, expected {self.variables}."
            )

        # Shift the point by the last vertex
        shift_point = point[self.variables] - self._vert_matrix[-1]
        shift_point = shift_point.tensor.reshape(-1, 1)

        # Shift the vertices by the last vertex
        shift_vert = (self._vert_matrix[:-1] - self._vert_matrix[-1]).T

        # Compute barycentric coordinates
        coords = torch.linalg.solve(shift_vert, shift_point)
        last_coord = 1.0 - torch.sum(coords)
        coords = torch.vstack([coords, last_coord])

        # If check_border is False -- use tolerance for numerical errors
        if not check_border:
            return torch.all(coords > 1e-6) & torch.all(coords < 1 - 1e-6)

        return torch.all(coords >= -1e-6) & torch.all(coords <= 1 + 1e-6)

    def update(self, domain):
        """
        Update the current domain by substituting the simplex vertices with
        those contained in ``domain``. Only domains of the same type can be used
        for update.

        :param SimplexDomain domain: The domain whose vertices are to be set
            into the current one.
        :raises TypeError: If the domain is not a :class:`SimplexDomain` object.
        :return: A new domain instance with the merged labels.
        :rtype: SimplexDomain
        """
        # Raise an error if the domain types do not match
        if not isinstance(domain, type(self)):
            raise TypeError(
                f"Cannot update domain of type {type(self)} "
                f"with domain of type {type(domain)}."
            )

        # Compute new vertex matrix
        vert_matrix = []
        for v in domain.vert_matrix:
            vert = v.reshape(1, -1)
            vert.labels = domain.variables
            vert_matrix.append(vert)

        # Replace geometry
        updated = deepcopy(self)
        updated.vert_matrix = vert_matrix

        return updated

    def sample(self, n, mode="random", variables="all"):
        """
        Sampling routine.

        :param int n: The number of samples to generate.
        :param str mode: The sampling method. Available modes: ``random`` for
            random sampling. Default is ``random``.
        :param variables: The list of variables to sample. If ``all``, all
            variables are sampled. Default is ``all``.
        :type variables: list[str] | str
        :raises AssertionError: If ``n`` is not a positive integer.
        :raises ValueError: If the sampling mode is invalid.
        :raises ValueError: If ``variables`` is neither ``all``, a string, nor a
            list/tuple of strings.
        :raises ValueError: If any of the specified variables is unknown.
        :return: The sampled points.
        :rtype: LabelTensor

        :Example:
            >>> simplex_domain = SimplexDomain(
                    [
                        LabelTensor(torch.tensor([[0, 0]]), labels=["x", "y"]),
                        LabelTensor(torch.tensor([[1, 1]]), labels=["x", "y"]),
                        LabelTensor(torch.tensor([[0, 1]]), labels=["x", "y"]),
                    ]
                )
            >>> simplex_domain.sample(n=5)
                LabelTensor([[0.0125, 0.0439],
                             [0.1346, 0.1950],
                             [0.8811, 0.9939],
                             [0.2722, 0.5535],
                             [0.4750, 0.7433]])
        """
        # Validate sampling settings
        variables = self._validate_sampling(n, mode, variables)

        # Extract vertex matrix for the requested variables
        vert_matrix = self._vert_matrix[variables].tensor

        # Sample barycentric coordinates using the Dirichlet distribution over
        # the simplex. This can be efficiently done by using samples obtained
        # via: -log(U(0,1)) ~ Exp(1) ~ Gamma(1, 1) ~ Dirichlet(1, ..., 1).
        coords = -torch.rand((n, vert_matrix.shape[0])).clamp_min(1e-12).log()

        # If only the surface is to be sampled
        if self._sample_surface:

            # Pick one face of the simplex at random for each point and set the
            # corresponding barycentric coordinate to zero.
            face_idx = torch.randint(0, vert_matrix.shape[0], (n,))
            coords.scatter_(1, face_idx.view(-1, 1), 0.0)

        # Normalize the coords
        coords = coords / coords.sum(dim=1, keepdim=True).clamp_min(1e-12)

        # Prepare output
        pts = (coords @ vert_matrix).as_subclass(LabelTensor)
        pts.labels = variables

        return pts[sorted(pts.labels)]

    def partial(self):
        """
        Return the boundary of the domain as a new domain object.

        :return: The boundary of the domain.
        :rtype: SimplexDomain
        """
        boundary = deepcopy(self)
        boundary.sample_surface = True

        return boundary

    @property
    def variables(self):
        """
        The list of variables of the domain.

        :return: The list of variables of the domain.
        :rtype: list[str]
        """
        return sorted(self._vert_matrix.labels)

    @property
    def domain_dict(self):
        """
        The dictionary representing the domain. For the simplex domain, the keys
        are of the form 'v0', 'v1', ..., 'vn', where each key corresponds to a
        vertex of the simplex.

        :return: The dictionary representing the domain.
        :rtype: dict
        """
        return {
            f"v{i}": self._vert_matrix[i]
            for i in range(self._vert_matrix.shape[0])
        }

    @property
    def range(self):
        """
        Return an empty dictionary since the simplex domain does not have range
        variables. Implemented to comply with the :class:`BaseDomain` interface.

        :return: The range variables of the domain.
        :rtype: dict
        """
        return {}

    @property
    def fixed(self):
        """
        Return an empty dictionary since the simplex domain does not have fixed
        variables. Implemented to comply with the :class:`BaseDomain` interface.

        :return: The fixed variables of the domain.
        :rtype: dict
        """
        return {}

    @property
    def sample_surface(self):
        """
        Whether only the surface of the simplex is considered part of the
        domain.

        :return: ``True`` if only the surface is considered part of the domain,
            ``False`` otherwise.
        :rtype: bool
        """
        return self._sample_surface

    @sample_surface.setter
    def sample_surface(self, value):
        """
        Setter for the sample_surface property.

        :param bool value: The new value for the sample_surface property.
        :raises ValueError: If ``value`` is not a boolean.
        """
        check_consistency(value, bool)
        self._sample_surface = value

    @property
    def vert_matrix(self):
        """
        The vertex matrix of the simplex.

        :return: The vertex matrix.
        :rtype: LabelTensor
        """
        return self._vert_matrix

    @vert_matrix.setter
    def vert_matrix(self, value):
        """
        Setter for the vertex matrix.

        :param LabelTensor value: The new vertex matrix.
        :raises ValueError: If any element of ``value`` is not a
            :class:`LabelTensor`.
        :raises TypeError: If ``value`` is not a list or tuple.
        :raises ValueError: If the labels of the vertices do not match.
        :raises ValueError: If the number of vertices is not equal to the
            dimension of the simplex plus one.
        """
        # Check consistency
        check_consistency(value, LabelTensor)
        if not isinstance(value, (list, tuple)):
            raise TypeError(
                "The simplex matrix must be a list or tuple of LabelTensor."
            )

        # Check that all labels match
        matrix_labels = value[0].labels
        if not all(vert.labels == matrix_labels for vert in value):
            raise ValueError("Labels of all vertices must match.")

        # Check dimensionality
        if len(value) != len(matrix_labels) + 1:
            raise ValueError(
                "An n-dimensional simplex needs n+1 vertices in R^n."
            )

        self._vert_matrix = LabelTensor.vstack(value).to(torch.float32)



"""Module for the Cartesian Domain."""

import torch
from .base_domain import BaseDomain
from .union import Union
from ..utils import torch_lhs, chebyshev_roots, check_consistency
from ..label_tensor import LabelTensor


class CartesianDomain(BaseDomain):
    """
    Implementation of the hypercube domain, obtained as the cartesian product of
    one-dimensional intervals.

    :Example:

        >>> cartesian_domain = CartesianDomain({'x': [0, 1], 'y': [0, 1]})
        >>> cartesian_domain = CartesianDomain({'x': [0, 1], 'y': 1.0})
    """

    def __init__(self, cartesian_dict):
        """
        Initialization of the :class:`CartesianDomain` class.

        :param dict cartesian_dict: A dictionary where the keys are the variable
            names and the values are the domain extrema. The domain extrema can
            be either a list or tuple with two elements or a single number. If
            the domain extrema is a single number, the variable is fixed to that
            value.
        :raises TypeError: If the cartesian dictionary is not a dictionary.
        :raises ValueError: If the cartesian dictionary contains variables with
            invalid ranges.
        :raises ValueError: If the cartesian dictionary contains values that are
            neither numbers nor lists/tuples of numbers of length 2.
        """
        # Initialization
        super().__init__(variables_dict=cartesian_dict)
        self._sample_modes = ("random", "grid", "chebyshev", "lh", "latin")

    def is_inside(self, point, check_border=False):
        """
        Check if a point is inside the domain.

        :param LabelTensor point: The point to check.
        :param bool check_border: If ``True``, the boundary is considered inside
            the domain. Default is ``False``.
        :raises ValueError: If ``point`` is not a :class:`LabelTensor`.
        :raises ValueError: If the labels of ``point`` differ from the variables
            of the domain.
        :return: Whether the point is inside the domain or not.
        :rtype: bool
        """
        # Checks on point
        check_consistency(point, LabelTensor)
        if set(self.variables) != set(point.labels):
            raise ValueError(
                "Point labels differ from domain's dictionary labels. "
                f"Got {sorted(point.labels)}, expected {self.variables}."
            )

        # Fixed variable checks
        fixed_check = all(
            (point.extract([k]) == v).all() for k, v in self._fixed.items()
        )

        # If there are no range variables, return fixed variable check
        if not self._range:
            return fixed_check

        # Ranged variable checks -- check_border True
        if check_border:
            range_check = all(
                (
                    (point.extract([k]) >= low) & (point.extract([k]) <= high)
                ).all()
                for k, (low, high) in self._range.items()
            )

        # Ranged variable checks -- check_border False
        else:
            range_check = all(
                ((point.extract([k]) > low) & (point.extract([k]) < high)).all()
                for k, (low, high) in self._range.items()
            )

        return fixed_check and range_check

    def sample(self, n, mode="random", variables="all"):
        """
        The sampling routine.

        :param int n: The number of samples to generate. See Note for reference.
        :param str mode: The sampling method. Available modes: ``random`` for
            random sampling; ``latin`` or ``lh`` for latin hypercube sampling;
            ``chebyshev`` for chebyshev sampling; ``grid`` for grid sampling.
            Default is ``random``.
        :param variables: The list of variables to sample. If ``all``, all
            variables are sampled. Default is ``all``.
        :type variables: list[str] | str
        :raises AssertionError: If ``n`` is not a positive integer.
        :raises ValueError: If the sampling mode is invalid.
        :raises ValueError: If ``variables`` is neither ``all``, a string, nor a
            list/tuple of strings.
        :raises ValueError: If any of the specified variables is unknown.
        :return: The sampled points.
        :rtype: LabelTensor

        .. note::
            When multiple variables are involved, the total number of sampled
            points may differ depending on the chosen ``mode``.
            If ``mode`` is ``grid`` or ``chebyshev``, points are sampled
            independently for each variable and then combined, resulting in a
            total number of points equal to ``n`` raised to the power of the
            number of variables. If ``mode`` is ``random``, ``lh`` or ``latin``,
            all variables are sampled together, and the total number of points
            remains ``n``.

        .. warning::
            The extrema of CartesianDomain are only sampled when using the
            ``grid`` mode.

        :Example:

            >>> cartesian_domain = CartesianDomain({'x': [0, 1], 'y': [0, 1]})
            >>> cartesian_domain.sample(n=3, mode='random')
                LabelTensor([[0.0108, 0.7643],
                             [0.4477, 0.8015],
                             [0.8735, 0.6349]])
            >>> cartesian_domain.sample(n=3, mode='grid')
                LabelTensor([[0.0000, 0.0000],
                             [0.5000, 0.0000],
                             [1.0000, 0.0000],
                             [0.0000, 0.5000],
                             [0.5000, 0.5000],
                             [1.0000, 0.5000],
                             [0.0000, 1.0000],
                             [0.5000, 1.0000],
                             [1.0000, 1.0000]])
        """
        # Validate sampling settings
        variables = self._validate_sampling(n, mode, variables)

        # Separate range and fixed variables
        range_vars = [v for v in variables if v in self._range]
        fixed_vars = [v for v in variables if v in self._fixed]

        # If there are no range variables, return fixed variables only
        if not range_vars:
            vals = [torch.full((n, 1), self._fixed[v]) for v in fixed_vars]
            result = torch.cat(vals, dim=1)
            result = result.as_subclass(LabelTensor)
            result.labels = fixed_vars
            return result

        # Create a tensor of bounds for the range variables
        bounds = torch.as_tensor([self._range[v] for v in range_vars])

        # Sample for mode random or latin hypercube
        if mode in {"random", "lh", "latin"}:
            pts = self._sample_range(n, mode, bounds)

        # Sample for mode grid or chebyshev
        else:
            grids = [
                self._sample_range(
                    n, mode, torch.as_tensor([self._range[v]])
                ).reshape(-1)
                for v in range_vars
            ]
            pts = torch.cartesian_prod(*grids).reshape(-1, len(grids))

        # Add fixed vars
        if fixed_vars:
            fixed_vals = [
                torch.full((pts.shape[0], 1), self._fixed[v])
                for v in fixed_vars
            ]
            pts = torch.cat([pts] + fixed_vals, dim=1)
            labels = range_vars + fixed_vars
        else:
            labels = range_vars

        # Create the result as a LabelTensor
        pts = pts.as_subclass(LabelTensor)
        pts.labels = labels

        return pts[sorted(pts.labels)]

    def _sample_range(self, n, mode, bounds):
        """
        Sample points and rescale to fit within the specified bounds.

        :param int n: The number of points to sample.
        :param str mode: The sampling method. Default is ``random``.
        :param torch.Tensor bounds: The bounds of the domain.
        :return: The rescaled sample points.
        :rtype: torch.Tensor
        """
        # Define a dictionary of sampling methods
        samplers = {
            "random": lambda: torch.rand(size=(n, bounds.shape[0])),
            "chebyshev": lambda: chebyshev_roots(n)
            .mul(0.5)
            .add(0.5)
            .reshape(-1, 1),
            "grid": lambda: torch.linspace(0, 1, n).reshape(-1, 1),
            "lh": lambda: torch_lhs(n, bounds.shape[0]),
            "latin": lambda: torch_lhs(n, bounds.shape[0]),
        }

        # Sample points in [0, 1]^d and rescale to the desired bounds
        pts = samplers[mode]()

        return pts * (bounds[:, 1] - bounds[:, 0]) + bounds[:, 0]

    def partial(self):
        """
        Return the boundary of the domain as a :class:`Union` object.

        :return: The boundary of the domain.
        :rtype: Union
        """
        faces = []

        # Iterate over ranged variables
        for var, (low, high) in self._range.items():

            # Fix the variable to its low value to get the lower face
            lower = CartesianDomain({**self._fixed, **self._range, var: low})

            # Fix the variable to its high value to get the upper face
            higher = CartesianDomain({**self._fixed, **self._range, var: high})

            faces.extend([lower, higher])

        return Union(faces)



"""Module for the Base class for domains."""

from copy import deepcopy
from abc import ABCMeta
from .domain_interface import DomainInterface
from ..utils import check_consistency, check_positive_integer


class BaseDomain(DomainInterface, metaclass=ABCMeta):
    """
    Base class for all geometric domains, implementing common functionality.

    All specific domain types should inherit from this class and implement the
    abstract methods of :class:`~pina.domain.domain_interface.DomainInterface`.

    This class is not meant to be instantiated directly.
    """

    def __init__(self, variables_dict=None):
        """
        Initialization of the :class:`BaseDomain` class.

        :param variables_dict: A dictionary where the keys are the variable
            names and the values are the domain extrema. The domain extrema can
            be either a list or tuple with two elements or a single number. If
            the domain extrema is a single number, the variable is fixed to that
            value.
        :type variables_dict: dict | None
        :raises TypeError: If the domain dictionary is not a dictionary.
        :raises ValueError: If the domain dictionary is empty.
        :raises ValueError: If the domain dictionary contains variables with
            invalid ranges.
        :raises ValueError: If the domain dictionary contains values that are
            neither numbers nor lists/tuples of numbers of length 2.
        """
        # Initialize fixed and ranged variables
        self._fixed = {}
        self._range = {}
        invalid = []

        # Skip checks if variables_dict is None -- SimplexDomain case
        if variables_dict is None:
            return

        # Check variables_dict is a dictionary
        if not isinstance(variables_dict, dict):
            raise TypeError(
                "variables_dict must be dict: {name: number | (low, high)}"
            )

        # Check variables_dict is not empty
        if not variables_dict:
            raise ValueError(
                "The dictionary defining the domain cannot be empty."
            )

        # Check consistency
        for v in variables_dict.values():
            check_consistency(v, (int, float))

        # Iterate over variables_dict items
        for k, v in variables_dict.items():

            # Fixed variables
            if isinstance(v, (int, float)):
                self._fixed[k] = v

            # Ranged variables
            elif isinstance(v, (list, tuple)) and len(v) == 2:
                low, high = v
                if low >= high:
                    raise ValueError(
                        f"Invalid range for variable '{k}': "
                        f"low ({low}) >= high ({high})"
                    )
                self._range[k] = (low, high)

            # Save invalid keys
            else:
                invalid.append(k)

        # Raise an error if there are invalid keys
        if invalid:
            raise ValueError(f"Invalid value(s) for key(s): {invalid}")

    def update(self, domain):
        """
        Update the current domain by adding the labels contained in ``domain``.
        Each new label introduces a new dimension. Only domains of the same type
        can be used for update.

        :param BaseDomain domain: The domain whose labels are to be merged
            into the current one.
        :raises TypeError: If the provided domain is not of the same type as
            the current one.
        :return: A new domain instance with the merged labels.
        :rtype: BaseDomain
        """
        # Raise an error if the domain types do not match
        if not isinstance(domain, type(self)):
            raise TypeError(
                f"Cannot update domain of type {type(self)} "
                f"with domain of type {type(domain)}."
            )

        # Update fixed and ranged variables
        updated = deepcopy(self)
        updated.fixed.update(domain.fixed)
        updated.range.update(domain.range)

        return updated

    def _validate_sampling(self, n, mode, variables):
        """
        Validate the sampling settings.

        :param int n: The number of samples to generate.
        :param str mode: The sampling method.
        :param variables: The list of variables to sample. If ``all``, all
            variables are sampled.
        :raises AssertionError: If ``n`` is not a positive integer.
        :raises ValueError: If the sampling mode is invalid.
        :raises ValueError: If ``variables`` is neither ``all``, a string, nor a
            list/tuple of strings.
        :raises ValueError: If any of the specified variables is unknown.
        :return: The validated list of variables to sample.
        :rtype: list[str]
        """
        # Validate n
        check_positive_integer(value=n, strict=True)

        # Validate mode
        if mode not in self.sample_modes:
            raise ValueError(
                f"Invalid sampling mode: {mode}. Available: {self.sample_modes}"
            )

        # Validate variables
        check_consistency(variables, str)
        if variables == "all":
            variables = self.variables
        elif isinstance(variables, str):
            variables = [variables]
        else:
            variables = list(dict.fromkeys(variables))

        # Check for unknown variables
        unknown = [v for v in variables if v not in self.variables]
        if unknown:
            raise ValueError(
                f"Unknown variable(s): {unknown}. Available: {self.variables}"
            )

        return sorted(variables)

    @property
    def sample_modes(self):
        """
        The list of available sampling modes.

        :return: The list of available sampling modes.
        :rtype: list[str]
        """
        return list(self._sample_modes)

    @property
    def variables(self):
        """
        The list of variables of the domain.

        :return: The list of variables of the domain.
        :rtype: list[str]
        """
        return sorted(list(self._fixed.keys()) + list(self._range.keys()))

    @property
    def domain_dict(self):
        """
        The dictionary representing the domain.

        :return: The dictionary representing the domain.
        :rtype: dict
        """
        return {**self._fixed, **self._range}

    @property
    def range(self):
        """
        The range variables of the domain.

        :return: The range variables of the domain.
        :rtype: dict
        """
        return self._range

    @property
    def fixed(self):
        """
        The fixed variables of the domain.

        :return: The fixed variables of the domain.
        :rtype: dict
        """
        return self._fixed



"""Module for the Difference operation."""

from .base_operation import BaseOperation
from ..label_tensor import LabelTensor
from ..utils import check_consistency


class Difference(BaseOperation):
    r"""
    Implementation of the difference operation defined on a list of domains.

    Given two sets :math:`A` and :math:`B`, define their difference as:

    .. math::

        A \setminus B = \{x \mid x \in A \land x \not\in B\}

    For multiple sets :math:`A_1, A_2, \ldots, A_n`, define their difference as
    the set of points that belong to the first set but not to any of the
    remaining sets.

    No check is performed to ensure that the resulting domain is non-empty.

    :Example:

        >>> cartesian1 = CartesianDomain({'x': [0, 1], 'y': [0, 1]})
        >>> cartesian2 = CartesianDomain({'x': [0, 1], 'y': [0.5, 1.5]})
        >>> difference = Difference([cartesian1, cartesian2])
    """

    def is_inside(self, point, check_border=False):
        """
        Check if a point is inside the difference of the domains.

        :param LabelTensor point: The point to check.
        :param bool check_border: If ``True``, the boundary is considered inside
            the domain. Default is ``False``.
        :raises ValueError: If ``point`` is not a :class:`LabelTensor`.
        :raises ValueError: If the labels of ``point`` differ from the variables
            of the domain.
        :return: Whether the point is inside the domain or not.
        :rtype: bool
        """
        # Checks on point
        check_consistency(point, LabelTensor)
        if set(self.variables) != set(point.labels):
            raise ValueError(
                "Point labels differ from domain's dictionary labels. "
                f"Got {sorted(point.labels)}, expected {self.variables}."
            )

        # Check if the point is inside the first geometry and not in any other
        inside_first = self.geometries[0].is_inside(point, check_border)
        inside_others = any(
            g.is_inside(point, check_border) for g in self.geometries[1:]
        )

        return inside_first and not inside_others

    def sample(self, n, mode="random", variables="all"):
        """
        The sampling routine.

        .. note::

            This sampling method relies on rejection sampling. Points are drawn
            from the individual geometries, and only those that lie exclusively
            within one geometry are kept. When the exclusion domain is small
            relative to the combined area of the input domains, the method may
            become highly inefficient.

        :param int n: The number of samples to generate.
        :param str mode: The sampling method. Default is ``random``.
        :param variables: The list of variables to sample. If ``all``, all
            variables are sampled. Default is ``all``.
        :type variables: list[str] | str
        :raises AssertionError: If ``n`` is not a positive integer.
        :raises ValueError: If the sampling mode is invalid.
        :raises ValueError: If ``variables`` is neither ``all``, a string, nor a
            list/tuple of strings.
        :raises ValueError: If any of the specified variables is unknown.
        :return: The sampled points.
        :rtype: LabelTensor
        """
        # Validate sampling settings
        variables = self._validate_sampling(n, mode, variables)

        # Allocate list for samples
        samples = []

        # Sample until we have enough points
        while len(samples) < n:

            # Sample a sufficiently large number of points
            batch_size = 2 * (n - len(samples))
            pts = self.geometries[0].sample(batch_size, mode)

            # Filter points inside the intersection
            for p in pts:
                p = p.reshape(1, -1)
                p.labels = pts.labels
                if self.is_inside(p):
                    samples.append(p[variables])
                    if len(samples) >= n:
                        break

        return LabelTensor.cat(samples, dim=0)

    def partial(self):
        """
        Return the boundary of the domain resulting from the operation.

        :raises NotImplementedError: The :meth:`partial` method is not
            implemented for difference domains. Please operate on the individual
            domains instead.
        """
        raise NotImplementedError(
            "The partial method is not implemented for difference domains. "
            "Please operate on the individual domains instead."
        )



"""Module to create and handle domains."""

__all__ = [
    "DomainInterface",
    "BaseDomain",
    "CartesianDomain",
    "EllipsoidDomain",
    "SimplexDomain",
    "OperationInterface",
    "Union",
    "Intersection",
    "Difference",
    "Exclusion",
]

from .domain_interface import DomainInterface
from .base_domain import BaseDomain
from .cartesian_domain import CartesianDomain
from .ellipsoid_domain import EllipsoidDomain
from .simplex_domain import SimplexDomain
from .operation_interface import OperationInterface
from .union import Union
from .intersection import Intersection
from .difference import Difference
from .exclusion import Exclusion



"""Module for the Ellipsoid Domain."""

from copy import deepcopy
import torch
from .base_domain import BaseDomain
from ..label_tensor import LabelTensor
from ..utils import check_consistency


class EllipsoidDomain(BaseDomain):
    """
    Implementation of the ellipsoid domain.

    .. seealso::

        **Original reference**: Dezert, Jean, and Musso, Christian.
        *An efficient method for generating points uniformly distributed
        in hyperellipsoids.*
        Proceedings of the Workshop on Estimation, Tracking and Fusion:
        A Tribute to Yaakov Bar-Shalom. 2001.

    :Example:

        >>> ellipsoid_domain = EllipsoidDomain({'x':[-1, 1], 'y':[-1, 1]})
        >>> ellipsoid_domain = EllipsoidDomain({'x':[-1, 1], 'y':1.0})
    """

    def __init__(self, ellipsoid_dict, sample_surface=False):
        """
        Initialization of the :class:`EllipsoidDomain` class.

        :param dict ellipsoid_dict: A dictionary where the keys are the variable
            names and the values are the domain extrema. The domain extrema can
            be either a list or tuple with two elements or a single number. If
            the domain extrema is a single number, the variable is fixed to that
            value.
        :param bool sample_surface: If ``True``, only the surface of the
            ellipsoid is considered part of the domain. Default is ``False``.
        :raises ValueError: If ``sample_surface`` is not a boolean.
        :raises TypeError: If the ellipsoid dictionary is not a dictionary.
        :raises ValueError: If the ellipsoid dictionary contains variables with
            invalid ranges.
        :raises ValueError: If the ellipsoid dictionary contains values that are
            neither numbers nor lists/tuples of numbers of length 2.
        """
        # Initialization
        super().__init__(variables_dict=ellipsoid_dict)
        self.sample_surface = sample_surface
        self._sample_modes = ("random",)
        self.compute_center_axes()

    def compute_center_axes(self):
        """
        Compute centers and axes for the ellipsoid.
        """
        if self._range:
            rng_vars = sorted(self._range.keys())
            vals = torch.tensor(
                [self._range[k] for k in rng_vars], dtype=torch.float
            )
            self._centers = LabelTensor(vals.mean(dim=1), rng_vars)
            self._axes = LabelTensor(
                (vals - self._centers.unsqueeze(1))[:, -1],
                rng_vars,
            )
        else:
            self._centers = None
            self._axes = None

    def is_inside(self, point, check_border=False):
        """
        Check if a point is inside the ellipsoid.

        :param LabelTensor point: The point to check.
        :param bool check_border: If ``True``, the boundary is considered inside
            the domain. Default is ``False``.
        :raises ValueError: If ``point`` is not a :class:`LabelTensor`.
        :raises ValueError: If the labels of ``point`` differ from the variables
            of the domain.
        :return: Whether the point is inside the domain or not.
        :rtype: bool
        """
        # Checks on point
        check_consistency(point, LabelTensor)
        if set(self.variables) != set(point.labels):
            raise ValueError(
                "Point labels differ from constructor dictionary labels. "
                f"Got {sorted(point.labels)}, expected {self.variables}."
            )

        # Fixed variable checks
        fixed_check = all(
            (point.extract([k]) == v).all() for k, v in self._fixed.items()
        )

        # If there are no range variables, return fixed variable check
        if not self._range:
            return fixed_check

        # Compute the equation defining the ellipsoid
        rng = sorted(self._range.keys())
        squared_axis = self._axes[rng].pow(2)
        delta = (point[rng] - self._centers[rng]).pow(2)
        eqn = torch.sum(delta / squared_axis) - 1.0

        # Range variable check on the surface
        if self._sample_surface:
            range_check = torch.allclose(eqn, torch.zeros_like(eqn))
            return fixed_check and range_check

        # Range variable check in the volume
        range_check = (eqn <= 0) if check_border else (eqn < 0)

        return fixed_check and range_check.item()

    def update(self, domain):
        """
        Update the current domain by adding the labels contained in ``domain``.
        Each new label introduces a new dimension. Only domains of the same type
        can be used for update.

        :param EllipsoidDomain domain: The domain whose labels are to be merged
            into the current one.
        :raises TypeError: If the provided domain is not of an instance of
            :class:`EllipsoidDomain`.
        :return: A new domain instance with the merged labels.
        :rtype: EllipsoidDomain
        """
        updated = super().update(domain)
        updated.compute_center_axes()

        return updated

    def sample(self, n, mode="random", variables="all"):
        """
        Sampling routine.

        :param int n: The number of samples to generate.
        :param str mode: The sampling method. Available modes: ``random`` for
            random sampling. Default is ``random``.
        :param variables: The list of variables to sample. If ``all``, all
            variables are sampled. Default is ``all``.
        :type variables: list[str] | str
        :raises AssertionError: If ``n`` is not a positive integer.
        :raises ValueError: If the sampling mode is invalid.
        :raises ValueError: If ``variables`` is neither ``all``, a string, nor a
            list/tuple of strings.
        :raises ValueError: If any of the specified variables is unknown.
        :return: The sampled points.
        :rtype: LabelTensor

        :Example:

            >>> ellipsoid_domain = EllipsoidDomain({'x':[0, 1], 'y':[0, 1]})
            >>> ellipsoid_domain.sample(n=5)
                LabelTensor([[0.7174, 0.5319],
                             [0.2713, 0.6518],
                             [0.1020, 0.4093],
                             [0.2102, 0.1353],
                             [0.4830, 0.1873]])
        """
        # Validate sampling settings
        variables = self._validate_sampling(n, mode, variables)

        # Separate range and fixed variables
        range_vars = [v for v in variables if v in self._range]
        fixed_vars = [v for v in variables if v in self._fixed]

        # If there are no range variables, return fixed variables only
        if not range_vars:
            vals = [torch.full((n, 1), self._fixed[v]) for v in fixed_vars]
            result = torch.cat(vals, dim=1)
            result = result.as_subclass(LabelTensor)
            result.labels = fixed_vars
            return result

        # Sample points
        pts = self._sample_range(n, range_vars)
        labels = range_vars

        # Add fixed vars
        if fixed_vars:
            fixed_vals = [
                torch.full((pts.shape[0], 1), self._fixed[v])
                for v in fixed_vars
            ]
            pts = torch.cat([pts] + fixed_vals, dim=1)
            labels = range_vars + fixed_vars

        # Prepare output
        pts = pts.as_subclass(LabelTensor)
        pts.labels = labels

        return pts[sorted(pts.labels)]

    def _sample_range(self, n, variables):
        """
        Sample points and rescale to fit within the specified bounds.

        :param int n: The number of points to sample.
        :param list[str] variables: variables whose samples must be rescaled.
        :return: The rescaled sample points.
        :rtype: torch.Tensor
        """
        # Extract the dimension
        dim = len(variables)

        # Extract centers and axes of the variables to sample
        centers = self._centers[variables]
        axes = self._axes[variables]

        # Find random directions on the unit sphere
        pts = torch.randn(size=(n, dim))
        norm = torch.linalg.vector_norm(pts, dim=1, keepdim=True)
        direction = pts / norm.clamp_min(1e-12)

        # Radius is set to one if sampling on the surface
        if self._sample_surface:
            radius = torch.ones((n, 1))

        # Otherwise, scale radius to lie within the sphere. Important: exponent
        # 1/dim is used to avoid shrinkage of the ellipsoid in higher dims.
        else:
            radius = torch.rand((n, 1)).pow(1.0 / dim)

        # Rescale the points to lie within the ellipsoid
        pts = direction * radius * axes + centers

        return pts

    def partial(self):
        """
        Return the boundary of the domain as a new domain object.

        :return: The boundary of the domain.
        :rtype: EllipsoidDomain
        """
        boundary = deepcopy(self)
        boundary.sample_surface = True

        return boundary

    @property
    def sample_surface(self):
        """
        Whether only the surface of the ellipsoid is considered part of the
        domain.

        :return: ``True`` if only the surface is considered part of the domain,
            ``False`` otherwise.
        :rtype: bool
        """
        return self._sample_surface

    @sample_surface.setter
    def sample_surface(self, value):
        """
        Setter for the sample_surface property.

        :param bool value: The new value for the sample_surface property.
        :raises ValueError: If ``value`` is not a boolean.
        """
        check_consistency(value, bool)
        self._sample_surface = value



"""Module for the Union operation."""

import random
from .base_operation import BaseOperation
from ..label_tensor import LabelTensor
from ..utils import check_consistency


class Union(BaseOperation):
    r"""
    Implementation of the union operation defined on a list of domains.

    Given multiple sets :math:`A_1, A_2, \ldots, A_n`, define their union as:

    .. math::

        \bigcup_{i=1}^{n} A_i = \{x \mid \exists i: x \in A_i \}

    :Example:

        >>> cartesian1 = CartesianDomain({'x': [0, 1], 'y': [0, 1]})
        >>> cartesian2 = CartesianDomain({'x': [0, 1], 'y': [1, 2]})
        >>> union = Union([cartesian1, cartesian2])
    """

    def is_inside(self, point, check_border=False):
        """
        Check if a point is inside the union of the domains.

        :param LabelTensor point: The point to check.
        :param bool check_border: If ``True``, the boundary is considered inside
            the domain. Default is ``False``.
        :raises ValueError: If ``point`` is not a :class:`LabelTensor`.
        :raises ValueError: If the labels of ``point`` differ from the variables
            of the domain.
        :return: Whether the point is inside the domain or not.
        :rtype: bool
        """
        # Checks on point
        check_consistency(point, LabelTensor)
        if set(self.variables) != set(point.labels):
            raise ValueError(
                "Point labels differ from domain's dictionary labels. "
                f"Got {sorted(point.labels)}, expected {self.variables}."
            )

        return any(g.is_inside(point, check_border) for g in self.geometries)

    def sample(self, n, mode="random", variables="all"):
        """
        The sampling routine.

        :param int n: The number of samples to generate.
        :param str mode: The sampling method. Default is ``random``.
        :param variables: The list of variables to sample. If ``all``, all
            variables are sampled. Default is ``all``.
        :type variables: list[str] | str
        :raises AssertionError: If ``n`` is not a positive integer.
        :raises ValueError: If the sampling mode is invalid.
        :raises ValueError: If ``variables`` is neither ``all``, a string, nor a
            list/tuple of strings.
        :raises ValueError: If any of the specified variables is unknown.
        :return: The sampled points.
        :rtype: LabelTensor
        """
        # Validate sampling settings
        variables = self._validate_sampling(n, mode, variables)

        # Compute number of points per geometry and remainder
        num_pts, remainder = divmod(n, len(self.geometries))

        # Shuffle indices
        shuffled_geometries = random.sample(
            range(len(self.geometries)), len(self.geometries)
        )

        # Precompute per-geometry allocations following the shuffled order
        alloc = [num_pts + (i < remainder) for i in range(len(self.geometries))]
        samples = []

        # Iterate over geometries in shuffled order
        for idx, gi in enumerate(shuffled_geometries):

            # If no points to allocate (possible if len(self.geometries) > n)
            if alloc[idx] == 0:
                continue

            # Sample points
            pts = self.geometries[gi].sample(alloc[idx], mode, variables)
            samples.append(pts)

        return LabelTensor.cat(samples, dim=0)

    def partial(self):
        """
        Return the boundary of the domain resulting from the operation.

        :raises NotImplementedError: The :meth:`partial` method is not
            implemented for union domains. Please operate on the individual
            domains instead.
        """
        raise NotImplementedError(
            "The partial method is not implemented for union domains. "
            "Please operate on the individual domains instead."
        )



"""Module for the Intersection operation."""

import random
from .base_operation import BaseOperation
from ..label_tensor import LabelTensor
from ..utils import check_consistency


class Intersection(BaseOperation):
    r"""
    Implementation of the intersection operation defined on a list of domains.

    Given multiple sets :math:`A_1, A_2, \ldots, A_n`, define their intersection
    as:

    .. math::

        \bigcap_{i=1}^{n} A_i = \{x \mid x \in A_i \forall i\}

    No check is performed to ensure that the resulting domain is non-empty.

    :Example:

        >>> cartesian1 = CartesianDomain({'x': [0, 1], 'y': [0, 1]})
        >>> cartesian2 = CartesianDomain({'x': [0, 1], 'y': [0.5, 1.5]})
        >>> intersection = Intersection([cartesian1, cartesian2])
    """

    def is_inside(self, point, check_border=False):
        """
        Check if a point is inside the intersection of the domains.

        :param LabelTensor point: The point to check.
        :param bool check_border: If ``True``, the boundary is considered inside
            the domain. Default is ``False``.
        :raises ValueError: If ``point`` is not a :class:`LabelTensor`.
        :raises ValueError: If the labels of ``point`` differ from the variables
            of the domain.
        :return: Whether the point is inside the domain or not.
        :rtype: bool
        """
        # Checks on point
        check_consistency(point, LabelTensor)
        if set(self.variables) != set(point.labels):
            raise ValueError(
                "Point labels differ from domain's dictionary labels. "
                f"Got {sorted(point.labels)}, expected {self.variables}."
            )

        return all(g.is_inside(point, check_border) for g in self.geometries)

    def sample(self, n, mode="random", variables="all"):
        """
        The sampling routine.

        .. note::

            This sampling method relies on rejection sampling. Points are drawn
            from the individual geometries, and only those that lie exclusively
            within one geometry are kept. When the exclusion domain is small
            relative to the combined area of the input domains, the method may
            become highly inefficient.

        :param int n: The number of samples to generate.
        :param str mode: The sampling method. Default is ``random``.
        :param variables: The list of variables to sample. If ``all``, all
            variables are sampled. Default is ``all``.
        :type variables: list[str] | str
        :raises AssertionError: If ``n`` is not a positive integer.
        :raises ValueError: If the sampling mode is invalid.
        :raises ValueError: If ``variables`` is neither ``all``, a string, nor a
            list/tuple of strings.
        :raises ValueError: If any of the specified variables is unknown.
        :return: The sampled points.
        :rtype: LabelTensor
        """
        # Validate sampling settings
        variables = self._validate_sampling(n, mode, variables)

        # Compute number of points per geometry and remainder
        num_pts, remainder = divmod(n, len(self.geometries))

        # Shuffle indices
        shuffled_geometries = random.sample(
            range(len(self.geometries)), len(self.geometries)
        )

        # Precompute per-geometry allocations following the shuffled order
        alloc = [num_pts + (i < remainder) for i in range(len(self.geometries))]
        samples = []

        # Iterate over geometries in shuffled order
        for idx, gi in enumerate(shuffled_geometries):

            # If no points to allocate (possible if len(self.geometries) > n)
            if alloc[idx] == 0:
                continue

            # Sampled points for the current geometry
            sampled_points = []

            # Sample until we have enough points
            while len(sampled_points) < alloc[idx]:

                # Sample a sufficiently large number of points
                batch_size = 2 * (alloc[idx] - len(sampled_points))
                pts = self.geometries[gi].sample(batch_size, mode)

                # Filter points inside the intersection
                for p in pts:
                    p = p.reshape(1, -1)
                    p.labels = pts.labels
                    if self.is_inside(p):
                        sampled_points.append(p[variables])
                        if len(sampled_points) >= alloc[idx]:
                            break

            # Sample points
            samples.append(LabelTensor.cat(sampled_points, dim=0))

        return LabelTensor.cat(samples, dim=0)

    def partial(self):
        """
        Return the boundary of the domain resulting from the operation.

        :raises NotImplementedError: The :meth:`partial` method is not
            implemented for intersection domains. Please operate on the
            individual domains instead.
        """
        raise NotImplementedError(
            "The partial method is not implemented for intersection domains. "
            "Please operate on the individual domains instead."
        )



"""Formulation of the acoustic wave problem."""

import torch
from ... import Condition
from ...problem import SpatialProblem, TimeDependentProblem
from ...utils import check_consistency
from ...domain import CartesianDomain
from ...equation import (
    Equation,
    SystemEquation,
    FixedValue,
    FixedGradient,
    AcousticWave,
)


def initial_condition(input_, output_):
    """
    Definition of the initial condition of the acoustic wave problem.

    :param LabelTensor input_: The input data of the problem.
    :param LabelTensor output_: The output data of the problem.
    :return: The residual of the initial condition.
    :rtype: LabelTensor
    """
    arg = torch.pi * input_["x"]
    return output_ - torch.sin(arg) - 0.5 * torch.sin(4 * arg)


class AcousticWaveProblem(TimeDependentProblem, SpatialProblem):
    r"""
    Implementation of the acoustic wave problem in the spatial interval
    :math:`[0, 1]` and temporal interval :math:`[0, 1]`.

    .. seealso::

        **Original reference**: Wang, Sifan, Xinling Yu, and
        Paris Perdikaris. *When and why PINNs fail to train:
        A neural tangent kernel perspective*. Journal of
        Computational Physics 449 (2022): 110768.
        DOI: `10.1016 <https://doi.org/10.1016/j.jcp.2021.110768>`_.

    :Example:

        >>> problem = AcousticWaveProblem(c=2.0)
    """

    output_variables = ["u"]
    spatial_domain = CartesianDomain({"x": [0, 1]})
    temporal_domain = CartesianDomain({"t": [0, 1]})

    domains = {
        "D": spatial_domain.update(temporal_domain),
        "t0": spatial_domain.update(CartesianDomain({"t": 0})),
        "boundary": spatial_domain.partial().update(temporal_domain),
    }

    conditions = {
        "boundary": Condition(domain="boundary", equation=FixedValue(0.0)),
        "t0": Condition(
            domain="t0",
            equation=SystemEquation(
                [Equation(initial_condition), FixedGradient(0.0, d="t")]
            ),
        ),
    }

    def __init__(self, c=2.0):
        """
        Initialization of the :class:`AcousticWaveProblem` class.

        :param c: The wave propagation speed. Default is 2.0.
        :type c: float | int
        """
        super().__init__()
        check_consistency(c, (float, int))
        self.c = c

        self.conditions["D"] = Condition(
            domain="D", equation=AcousticWave(self.c)
        )

    def solution(self, pts):
        """
        Implementation of the analytical solution of the acoustic wave problem.

        :param LabelTensor pts: Points where the solution is evaluated.
        :return: The analytical solution of the acoustic wave problem.
        :rtype: LabelTensor
        """
        arg_x = torch.pi * pts["x"]
        arg_t = self.c * torch.pi * pts["t"]
        term1 = torch.sin(arg_x) * torch.cos(arg_t)
        term2 = 0.5 * torch.sin(4 * arg_x) * torch.cos(4 * arg_t)
        return term1 + term2



"""Module for the R3Refinement callback."""

import torch
from .refinement_interface import RefinementInterface
from ...label_tensor import LabelTensor
from ...utils import check_consistency
from ...loss import LossInterface


class R3Refinement(RefinementInterface):
    """
    PINA Implementation of the R3 Refinement Callback.

    This callback implements the R3 (Retain-Resample-Release) routine for
    sampling new points based on adaptive search.
    The algorithm incrementally accumulates collocation points in regions
    of high PDE residuals, and releases those with low residuals.
    Points are sampled uniformly in all regions where sampling is needed.

    .. seealso::

        Original Reference: Daw, Arka, et al. *Mitigating Propagation
        Failures in Physics-informed Neural Networks
        using Retain-Resample-Release (R3) Sampling. (2023)*.
        DOI: `10.48550/arXiv.2207.02338
        <https://doi.org/10.48550/arXiv.2207.02338>`_

    :Example:

    >>> r3_callback = R3Refinement(sample_every=5)
    """

    def __init__(
        self,
        sample_every,
        residual_loss=torch.nn.L1Loss,
        condition_to_update=None,
    ):
        """
        Initialization of the :class:`R3Refinement` callback.

        :param int sample_every: The sampling frequency.
        :param loss: The loss function to compute the residuals.
            Default is :class:`~torch.nn.L1Loss`.
        :type loss: LossInterface | :class:`~torch.nn.modules.loss._Loss`
        :param condition_to_update: The conditions to update during the
            refinement process. If None, all conditions will be updated.
            Default is None.
        :type condition_to_update: list(str) | tuple(str) | str
        :raises ValueError: If the condition_to_update is neither a string nor
            an iterable of strings.
        :raises TypeError: If the residual_loss is not a subclass of
            :class:`~torch.nn.Module`.
        """
        super().__init__(sample_every, condition_to_update)

        # Check consistency
        check_consistency(
            residual_loss,
            (LossInterface, torch.nn.modules.loss._Loss),
            subclass=True,
        )

        # Save loss function
        self.loss_fn = residual_loss(reduction="none")

    def sample(self, current_points, condition_name, solver):
        """
        Sample new points based on the R3 refinement strategy.

        :param current_points: The current points in the domain.
        :type current_points: LabelTensor | torch.Tensor
        :param str condition_name: The name of the condition to update.
        :param PINNInterface solver: The solver using this callback.
        :return: The new samples generated by the R3 strategy.
        :rtype: LabelTensor
        """
        # Retrieve condition and current points
        device = solver.trainer.strategy.root_device
        condition = solver.problem.conditions[condition_name]
        current_points = current_points.to(device).requires_grad_(True)

        # Compute residuals for the given condition (averaged over all fields)
        target = solver.compute_residual(current_points, condition.equation)
        residuals = self.loss_fn(target, torch.zeros_like(target)).mean(
            dim=tuple(range(1, target.ndim))
        )

        # Retrieve domain and initial population size
        domain_name = solver.problem.conditions[condition_name].domain
        domain = solver.problem.domains[domain_name]
        num_old_points = self.initial_population_size[condition_name]

        # Select points with residual above the mean
        mask = (residuals > residuals.mean()).flatten()
        if mask.any():
            high_residual_pts = current_points[mask]
            high_residual_pts.labels = current_points.labels
            samples = domain.sample(num_old_points - len(high_residual_pts))
            return LabelTensor.cat([high_residual_pts, samples.to(device)])

        return domain.sample(num_old_points, "random")



"""Module for the Metric Tracker."""

import copy
import torch
from lightning.pytorch.callbacks import Callback


class MetricTracker(Callback):
    """
    Lightning Callback for Metric Tracking.
    """

    def __init__(self, metrics_to_track=None):
        """
        Tracks specified metrics during training.

        :param metrics_to_track: List of metrics to track.
            Defaults to train loss.
        :type metrics_to_track: list[str], optional
        """
        super().__init__()
        self._collection = []
        # Default to tracking 'train_loss' if not specified
        self.metrics_to_track = metrics_to_track

    def setup(self, trainer, pl_module, stage):
        """
        Called when fit, validate, test, predict, or tune begins.

        :param Trainer trainer: A :class:`~pina.trainer.Trainer` instance.
        :param SolverInterface pl_module: A
            :class:`~pina.solver.solver.SolverInterface` instance.
        :param str stage: Either 'fit', 'test' or 'predict'.
        """
        if self.metrics_to_track is None and trainer.batch_size is None:
            self.metrics_to_track = ["train_loss"]
        elif self.metrics_to_track is None:
            self.metrics_to_track = ["train_loss_epoch"]
        return super().setup(trainer, pl_module, stage)

    def on_train_epoch_end(self, trainer, pl_module):
        """
        Collect and track metrics at the end of each training epoch.

        :param trainer: The trainer object managing the training process.
        :type trainer: pytorch_lightning.Trainer
        :param pl_module: The model being trained (not used here).
        """
        # Track metrics after the first epoch onwards
        if trainer.current_epoch > 0:
            # Append only the tracked metrics to avoid unnecessary data
            tracked_metrics = {
                k: v
                for k, v in trainer.logged_metrics.items()
                if k in self.metrics_to_track
            }
            self._collection.append(copy.deepcopy(tracked_metrics))

    @property
    def metrics(self):
        """
        Aggregate collected metrics over all epochs.

        :return: A dictionary containing aggregated metric values.
        :rtype: dict
        """
        if not self._collection:
            return {}

        # Get intersection of keys across all collected dictionaries
        common_keys = set(self._collection[0]).intersection(
            *self._collection[1:]
        )

        # Stack the metric values for common keys and return
        return {
            k: torch.stack([dic[k] for dic in self._collection])
            for k in common_keys
            if k in self.metrics_to_track
        }



"""Module for the Pina Callbacks."""

__all__ = [
    "SwitchOptimizer",
    "SwitchScheduler",
    "NormalizerDataCallback",
    "PINAProgressBar",
    "MetricTracker",
    "R3Refinement",
]

from .optim.switch_optimizer import SwitchOptimizer
from .optim.switch_scheduler import SwitchScheduler
from .processing.normalizer_data_callback import NormalizerDataCallback
from .processing.pina_progress_bar import PINAProgressBar
from .processing.metric_tracker import MetricTracker
from .refinement import R3Refinement



"""Module for all set-based operations Base class."""

from copy import deepcopy
from abc import ABCMeta
from .operation_interface import OperationInterface
from .base_domain import BaseDomain
from ..utils import check_consistency


class BaseOperation(OperationInterface, BaseDomain, metaclass=ABCMeta):
    """
    Base class for all set operation defined on geometric domains, implementing
    common functionality.

    All specific operation types should inherit from this class and implement
    the abstract methods defined in both the following interfaces:
    :class:`~pina.domain.operation_interface.OperationInterface`, and
    :class:`~pina.domain.domain_interface.DomainInterface`.

    This class is not meant to be instantiated directly.
    """

    def __init__(self, geometries):
        """
        Initialization of the :class:`OperationInterface` class.

        :param geometries: The list of domains on which to perform the set
            operation.
        :type geometries: list[BaseDomain] | tuple[BaseDomain]
        :raises TypeError: If geometries is neither a list nor a tuple.
        :raises ValueError: If geometries elements are not instances of
            :class:`~pina.domain.base_domain.BaseDomain`.
        :raises NotImplementedError: If the dimensions of the geometries are not
            consistent.
        """
        super().__init__()
        self.geometries = geometries

    def update(self, domain):
        """
        Update the domain resulting from the operation.

        :param DomainInterface domain: The domain whose labels are to be merged
            into the current one.
        :raises NotImplementedError: If the geometries involved in the operation
            are of different types.
        :raises TypeError: If the passed domain is not of the same type of all
            the geometries involved in the operation.
        :return: A new domain instance with the merged labels.
        :rtype: BaseOperation
        """
        # Check all geometries are of the same type
        domain_type = type(self.geometries[0])
        if not all(isinstance(g, domain_type) for g in self.geometries):
            raise NotImplementedError(
                f"The {self.__class__.__name__} of geometries of different"
                " types does not support the update operation. All geometries"
                " must be of the same type."
            )

        # Check domain type consistency
        if not isinstance(domain, domain_type):
            raise TypeError(
                f"Cannot update the {self.__class__.__name__} of domains of"
                f" type {domain_type} with domain of type {type(domain)}."
            )

        # Update each geometry
        updated = deepcopy(self)
        updated.geometries = [geom.update(domain) for geom in self.geometries]

        return updated

    @property
    def sample_modes(self):
        """
        The list of available sampling modes.

        :return: The list of available sampling modes.
        :rtype: list[str]
        """
        return list(
            set.intersection(
                *map(set, [g.sample_modes for g in self.geometries])
            )
        )

    @property
    def variables(self):
        """
        The list of variables of the domain.

        :return: The list of variables of the domain.
        :rtype: list[str]
        """
        return sorted({v for g in self.geometries for v in g.variables})

    @property
    def domain_dict(self):
        """
        Returns a dictionary representation of the operation domain.

        :return: The dictionary representation of the operation domain.
        :rtype: dict
        """
        return {
            "type": self.__class__.__name__,
            "geometries": [geom.domain_dict for geom in self.geometries],
        }

    @property
    def geometries(self):
        """
        The domains on which to perform the set operation.

        :return: The domains on which to perform the set operation.
        :rtype: list[BaseDomain]
        """
        return self._geometries

    @property
    def range(self):
        """
        The range variables of each geometry.

        :return: The range variables of each geometry.
        :rtype: dict
        """
        return {f"geometry_{i}": g.range for i, g in enumerate(self.geometries)}

    @property
    def fixed(self):
        """
        The fixed variables of each geometry.

        :return: The fixed variables of each geometry.
        :rtype: dict
        """
        return {f"geometry_{i}": g.fixed for i, g in enumerate(self.geometries)}

    @geometries.setter
    def geometries(self, values):
        """
        Setter for the ``geometries`` property.

        :param values: The geometries to be set.
        :type values: list[BaseDomain] | tuple[BaseDomain]
        :raises TypeError: If values is neither a list nor a tuple.
        :raises ValueError: If values elements are not instances of
            :class:`~pina.domain.base_domain.BaseDomain`.
        :raises NotImplementedError: If the dimensions of the geometries are not
            consistent.
        """
        # Check geometries are list or tuple
        if not isinstance(values, (list, tuple)):
            raise TypeError(
                "geometries must be either a list or a tuple of BaseDomain."
            )

        # Check consistency
        check_consistency(values, (BaseDomain, BaseOperation))

        # Check geometries
        for v in values:
            if v.variables != values[0].variables:
                raise NotImplementedError(
                    f"The {self.__class__.__name__} of geometries living in "
                    "different ambient spaces is not well-defined. "
                    "All geometries must share the same dimensions and labels."
                )

        self._geometries = values



"""Module for the SwitchScheduler callback."""

from lightning.pytorch.callbacks import Callback
from ...optim import TorchScheduler
from ...utils import check_consistency, check_positive_integer


class SwitchScheduler(Callback):
    """
    Callback to switch scheduler during training.
    """

    def __init__(self, new_schedulers, epoch_switch):
        """
        This callback allows switching between different schedulers during
        training, enabling the exploration of multiple optimization strategies
        without interrupting the training process.

        :param new_schedulers: The scheduler or list of schedulers to switch to.
            Use a single scheduler for single-model solvers, or a list of
            schedulers when working with multiple models.
        :type new_schedulers: pina.optim.TorchScheduler |
            list[pina.optim.TorchScheduler]
        :param int epoch_switch: The epoch at which the scheduler switch occurs.
        :raise AssertionError: If epoch_switch is less than 1.
        :raise ValueError: If each scheduler in ``new_schedulers`` is not an
            instance of :class:`pina.optim.TorchScheduler`.

        Example:
            >>> scheduler = TorchScheduler(
            >>>     torch.optim.lr_scheduler.StepLR, step_size=5
            >>> )
            >>> switch_callback = SwitchScheduler(
            >>>     new_schedulers=scheduler, epoch_switch=10
            >>> )
        """
        super().__init__()

        # Check if epoch_switch is greater than 1
        check_positive_integer(epoch_switch - 1, strict=True)

        # If new_schedulers is not a list, convert it to a list
        if not isinstance(new_schedulers, list):
            new_schedulers = [new_schedulers]

        # Check consistency
        for scheduler in new_schedulers:
            check_consistency(scheduler, TorchScheduler)

        # Store the new schedulers and epoch switch
        self._new_schedulers = new_schedulers
        self._epoch_switch = epoch_switch

    def on_train_epoch_start(self, trainer, __):
        """
        Switch the scheduler at the start of the specified training epoch.

        :param lightning.pytorch.Trainer trainer: The trainer object managing
            the training process.
        :param __: Placeholder argument (not used).
        """
        # Check if the current epoch matches the switch epoch
        if trainer.current_epoch == self._epoch_switch:
            schedulers = []

            # Hook the new schedulers to the model parameters
            for idx, scheduler in enumerate(self._new_schedulers):
                scheduler.hook(trainer.solver._pina_optimizers[idx])
                schedulers.append(scheduler)

                # Update the trainer's scheduler configs
                trainer.lr_scheduler_configs[idx].scheduler = scheduler.instance

            # Update the solver's schedulers
            trainer.solver._pina_schedulers = schedulers



"""Module for the Processing Callbacks."""

from lightning.pytorch.callbacks import TQDMProgressBar
from lightning.pytorch.callbacks.progress.progress_bar import (
    get_standard_metrics,
)
from pina.utils import check_consistency


class PINAProgressBar(TQDMProgressBar):
    """
    PINA Implementation of a Lightning Callback for enriching the progress bar.
    """

    BAR_FORMAT = (
        "{l_bar}{bar}| {n_fmt}/{total_fmt} [{elapsed}<{remaining}, "
        "{rate_noinv_fmt}{postfix}]"
    )

    def __init__(self, metrics="val", **kwargs):
        """
        This class enables the display of only relevant metrics during training.

        :param metrics: Logged metrics to be shown during the training.
            Must be a subset of the conditions keys defined in
            :obj:`pina.condition.Condition`.
        :type metrics: str | list(str) | tuple(str)

        :Keyword Arguments:
            The additional keyword arguments specify the progress bar and can be
            choosen from the `pytorch-lightning TQDMProgressBar API
            <https://lightning.ai/docs/pytorch/stable/_modules/lightning/pytorch/callback/progress/tqdm_progress.html#TQDMProgressBar>`_

        Example:
            >>> pbar = PINAProgressBar(['mean'])
            >>> # ... Perform training ...
            >>> trainer = Trainer(solver, callbacks=[pbar])
        """
        super().__init__(**kwargs)
        # check consistency
        if not isinstance(metrics, (list, tuple)):
            metrics = [metrics]
        check_consistency(metrics, str)
        self._sorted_metrics = metrics

    def get_metrics(self, trainer, pl_module):
        r"""Combine progress bar metrics collected from the trainer with
        standard metrics from get_standard_metrics.
        Override this method to customize the items shown in the progress bar.
        The progress bar metrics are sorted according to ``metrics``.

        Here is an example of how to override the defaults:

        .. code-block:: python

            def get_metrics(self, trainer, model):
                # don't show the version number
                items = super().get_metrics(trainer, model)
                items.pop("v_num", None)
                return items

        :return: Dictionary with the items to be displayed in the progress bar.
        :rtype: tuple(dict)
        """
        standard_metrics = get_standard_metrics(trainer)
        pbar_metrics = trainer.progress_bar_metrics
        if pbar_metrics:
            pbar_metrics = {
                key: pbar_metrics[key]
                for key in pbar_metrics
                if key in self._sorted_metrics
            }
        return {**standard_metrics, **pbar_metrics}

    def setup(self, trainer, pl_module, stage):
        """
        Check that the initialized metrics are available and correctly logged.

        :param trainer: The trainer object managing the training process.
        :type trainer: pytorch_lightning.Trainer
        :param pl_module: Placeholder argument.
        """
        # Check if all keys in sort_keys are present in the dictionary
        for key in self._sorted_metrics:
            if (
                key not in trainer.solver.problem.conditions.keys()
                and key != "train"
                and key != "val"
            ):
                raise KeyError(f"Key '{key}' is not present in the dictionary")
        # add the loss pedix
        if trainer.batch_size is not None:
            pedix = "_loss_epoch"
        else:
            pedix = "_loss"
        self._sorted_metrics = [
            metric + pedix for metric in self._sorted_metrics
        ]
        return super().setup(trainer, pl_module, stage)



"""Module for the Operation Interface."""

from abc import ABCMeta, abstractmethod
from .domain_interface import DomainInterface


class OperationInterface(DomainInterface, metaclass=ABCMeta):
    """
    Abstract interface for all set operations defined on geometric domains.
    """

    @property
    @abstractmethod
    def geometries(self):
        """
        The list of domains on which to perform the set operation.

        :return: The list of domains on which to perform the set operation.
        :rtype: list[BaseDomain]
        """

    @geometries.setter
    @abstractmethod
    def geometries(self, values):
        """
        Setter for the ``geometries`` property.

        :param values: The geometries to be set.
        :type values: list[BaseDomain] | tuple[BaseDomain]
        """



"""Module for the Domain Interface."""

from abc import ABCMeta, abstractmethod


class DomainInterface(metaclass=ABCMeta):
    """
    Abstract interface for all geometric domains.
    """

    @abstractmethod
    def is_inside(self, point, check_border):
        """
        Check if a point is inside the domain.

        :param LabelTensor point: The point to check.
        :param bool check_border: If ``True``, the boundary is considered inside
            the domain.
        :return: Whether the point is inside the domain or not.
        :rtype: bool
        """

    @abstractmethod
    def update(self, domain):
        """
        Update the current domain by adding the labels contained in ``domain``.
        Each new label introduces a new dimension. Only domains of the same type
        can be used for update.

        :param BaseDomain domain: The domain whose labels are to be merged into
            the current one.
        :return: A new domain instance with the merged labels.
        :rtype: DomainInterface
        """

    @abstractmethod
    def sample(self, n, mode, variables):
        """
        The sampling routine.

        :param int n: The number of samples to generate.
        :param str mode: The sampling method.
        :param list[str] variables: The list of variables to sample.
        :return: The sampled points.
        :rtype: LabelTensor
        """

    @abstractmethod
    def partial(self):
        """
        Return the boundary of the domain as a new domain object.

        :return: The boundary of the domain.
        :rtype: DomainInterface
        """

    @property
    @abstractmethod
    def sample_modes(self):
        """
        The list of available sampling modes.

        :return: The list of available sampling modes.
        :rtype: list[str]
        """

    @property
    @abstractmethod
    def variables(self):
        """
        The list of variables of the domain.

        :return: The list of variables of the domain.
        :rtype: list[str]
        """

    @property
    @abstractmethod
    def domain_dict(self):
        """
        The dictionary representing the domain.

        :return: The dictionary representing the domain.
        :rtype: dict
        """

    @property
    @abstractmethod
    def range(self):
        """
        The range variables of the domain.

        :return: The range variables of the domain.
        :rtype: dict
        """

    @property
    @abstractmethod
    def fixed(self):
        """
        The fixed variables of the domain.

        :return: The fixed variables of the domain.
        :rtype: dict
        """
